; g r o w o r l d   g a m e   p r o t o t y p e s
; [ copyright (c) 2010 foam vzw : gpl v3 ]

; The main script, and the menu system thingy

(define font "textures/zeimusu.ttf")
(define text-pos (vector -9.5 3 0))
(define plant-pos (vector 5 -4 0))
(define plant-col (vector 0.7 0.8 0.2))

(define (make-text txt outline)    
    (let ((root (build-locator)))
        (with-state
            (identity)
            (parent root)
            (hint-unlit)
            (if outline
                (colour (vector 0.8 1 0.9))
                (colour (vector 0.0 0.4 0.3)))        
            
            (build-type font txt)
            (when outline
                (wire-colour 0)
                (hint-none)
                ;(wire-opacity 0.5)
                (hint-wire)
                (line-width 4)
                (translate (vector 0 0 -0.01))
                (build-type font txt)) root)))

(define (searchpath-add path)
    (when
        (not (foldl
                (lambda (p r)
                    (or r (string=? path p)))
                #f
                (get-searchpaths)))
        (printf "adding searchpath ~a~n" path)
        (set-searchpaths (cons path (get-searchpaths)))))

(define (get-line-from-mouse)
    (let* ((ndcpos (vector (* (- (/ (mouse-x) (vx (get-screen-size))) 0.5) 2) ; 2
                    (* (- (- (/ (mouse-y) (vy (get-screen-size))) 0.5)) 1.5) -1)) ;1.5
            (scrpos2 (vtransform (vmul ndcpos 10) (minverse (get-camera-transform))))
            (scrpos (vtransform ndcpos (minverse (get-camera-transform)))))
        (list scrpos scrpos2)))

(define (choose l)
    (list-ref l (random (length l))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-struct plant (root shape widths (start-time #:mutable) (started #:mutable) (seed #:mutable)))

(define (build-plant size r d w off)
    (make-plant
        (with-state 
            (translate (vector 1000 0 0))
            (hint-unlit)
            (build-ribbon size))
        (let ((p (vector 0 0 0)))
            (build-list size 
                (lambda (i) 
                    (set! p (vadd p d))
                    (set! d (vtransform d (mrotate (vector 0 0 (* r (+ off (* 1 (sin (* i w)))))))))
                    (set! d (vmul d 0.99))
                    p)))
        (build-list size
            (lambda (i)
                (let ((t (/ i size)))
                    (* 5 t t (* 0.3 (+ 1 (cos (* t 3.141))))))))
        (time)
        #f
        (* (rndf) 10)))

(define (reset-plant plant pos col tex)
    (with-primitive (plant-root plant)
        (identity)
        (colour col)
        (texture (load-texture tex))       
        ;(shader-set! (list "ColourMap" 0))
        (translate pos)
        (pdata-map!
            (lambda (p)
                (vector 0 0 0))
            "p")
        (pdata-index-map!
            (lambda (i w)
                (list-ref (plant-widths plant) i))
            "w"))
    (set-plant-start-time! plant (time))
    (set-plant-started! plant #t))

(define (plant-update plant)
    (let ((t (* (- (time) (plant-start-time plant)) 2.5)))
        (with-primitive (plant-root plant)
            ;(shader-set! (list "Time" (* 10 (time))))
            (rotate (vector 0 0 (* 0.05 (sin (+ (plant-seed plant)
                                (- (time) (plant-start-time plant)))))))
            
            (when (< t (pdata-size))
                (pdata-index-map!
                    (lambda (i p)
                        (let ((t (- t i)))
                            (if (> t 0) 
                                (vmix p (list-ref (plant-shape plant) (inexact->exact (floor t))) 0.9)
                                p)))
                    "p")))))


(define (plant-destroy plant)
    (destroy (plant-root plant)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-struct slippery (string objs follow rev))

(define (build-slippery str follow rev) 
    (make-slippery
        str
        (map 
            (lambda (ch)               
                (make-text (string ch) #t))
            (string->list str))
        follow
        rev))

(define (f n)
  (exact->inexact n))

(define (slippery-update s)
    (for ((i (in-range 0 (length (slippery-objs s)))))         
        (let* ((rev (slippery-rev s))
               (pos (* (if rev
                           (- 1 (/ i (length (slippery-objs s))))
                           (/ i (length (slippery-objs s))))
                       (with-primitive (slippery-follow s) (- (pdata-size) 5))))
               (from (with-primitive (slippery-follow s) 
                                     (vtransform (pdata-ref "p" pos) (get-transform))))
               (to (with-primitive (slippery-follow s)
                                   (vtransform (pdata-ref "p"
                                                          (if rev
                                                              (- pos 1)
                                                              (+ pos 1)))
                                               (get-transform)))))
          (with-primitive (list-ref (slippery-objs s) i)
                          (identity)                    
                          (translate (vector 0 0 0.1))
                          (translate from)
                          (concat (maim (vsub from to) (vector 0 0 1)))
                          (rotate (vector 0 180 0))
                          (scale 0.2)
                          (translate (vector 0 -0.5 0))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-game name dir script info plant-params)
    (list name dir script info plant-params))

(define (game-name game) (list-ref game 0))
(define (game-dir game) (list-ref game 1))
(define (game-script game) (list-ref game 2))
(define (game-info game) (list-ref game 3))
(define (game-plant-params game) (list-ref game 4))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-struct menu-item (game plant slippery))

(define (build-menu-item game)
    (searchpath-add (game-dir game))
    (let ((plant (apply build-plant (game-plant-params game))))
        (reset-plant plant plant-pos plant-col 
            (choose (list "textures/leaf-01.png" 
                    "textures/leaf-02.png"
                    "textures/leaf-03.png")))
        (make-menu-item game plant
            (build-slippery (game-name game)
                (plant-root plant)
                (< (cadr (game-plant-params game)) 0)))))

(define (menu-item-dist menu-item pos)
    (foldl
        (lambda (obj r)
            (let ((d (vdist pos (with-primitive
                                obj
                                (vtransform (vector 0 0 0) (get-transform))))))
                (if (< d r) d r)))
        9999999
        (slippery-objs (menu-item-slippery menu-item))))

(define (menu-item-highlight menu-item)
    (for-each
        (lambda (obj)
            (with-primitive obj
                ;(translate (vmul (vector (crndf) (crndf) 0) 0))
                (rotate (vector 0 0 (* 15 (crndf))))))
        (slippery-objs (menu-item-slippery menu-item))))


(define (menu-item-build-abouttext menu-item)
    ; do linebreaks
    (let* ((t
                (foldl
                    (lambda (ch r)
                        (if (and (char=? ch #\ ) (> (string-length (car r)) 30))
                            (list "" (append (cadr r) (list (car r))))
                            (list (string-append (car r) (string ch))
                                (cadr r))))                         
                    (list "" '())
                    (string->list
                        (game-info (menu-item-game menu-item)))))
            (txt-list (append (cadr t) (list (car t))))
            (root (with-state
                    (translate text-pos)
                    (build-locator))))
        (with-state
            (scale 0.4)
            (parent root)
            (make-text (game-name (menu-item-game menu-item)) #f))
        (with-state
            (hint-unlit)
            (scale 0.2)
            (colour 0)
            (parent root)
            (translate (vector 0 -6 0))
            (for-each
                (lambda (line)        
                    (make-text line #f)
                    (translate (vector 0 -4 0)))
                txt-list))
        root))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-struct menu (items
        (about-root #:mutable)
        (current-about #:mutable)
        (closest #:mutable)))

(define (build-menu games)
    (make-menu
        (map
            (lambda (game)
                (build-menu-item game))
            games)
        0 "" #f))

(define (menu-check menu)
    (let* ((line (get-line-from-mouse))
            (pos (cadr line))
            (closest             
                (foldl
                    (lambda (menu-item r)
                        (let ((d (menu-item-dist menu-item pos)))
                            (if (< d (car r))
                                (list d menu-item)
                                r)))
                    (list 9999999 #f)
                    (menu-items menu))))
        ;(with-state (translate pos) (draw-sphere))
        (cond ((< (car closest) 1)      
                (menu-item-highlight (cadr closest))
                (let ((name (game-name (menu-item-game (cadr closest)))))
                    (when (not (string=? (menu-current-about menu) name))
                        (set-menu-closest! menu (cadr closest))
                        (set-menu-current-about! menu name)
                        (destroy (menu-about-root menu))
                        (set-menu-about-root!
                            menu
                            (with-state (menu-item-build-abouttext (cadr closest)))))))
            (else
                (when (not (string=? (menu-current-about menu) ""))
                    (set-menu-closest! menu #f)
                    (set-menu-current-about! menu "")             
                    (destroy (menu-about-root menu))
                    (set-menu-about-root! menu (build-locator)))))))

(define (menu-animate menu)
    (for-each
        (lambda (menu-item)
            (plant-update (menu-item-plant menu-item))
            (slippery-update (menu-item-slippery menu-item)))
        (menu-items menu))
    (menu-check menu)
    #;(for-each
        (lambda (menu-item)
            (menu-item-unhighlight menu-item))
        (menu-items menu)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(clear)
(clear-colour (vector 0.2 0.5 0.4))

(define splash2
    (with-state 
        (texture (load-texture "textures/intro-bg2.png"))
        (hint-unlit)
        (scale (vmul (vector 3 2.25 1) 12.5))
        (translate (vector 0 0 -0.7))
        (build-plane)))

#;(define splash
    (with-state 
        (texture (load-texture "textures/foam.png"))
        (hint-unlit)
        ;         (colour 0)
        (scale (vmul (vector 3 2 1) 14))
        (translate (vector 0 0 1))
        ;       (rotate (vector 90 0 0))
        (build-plane)))

(set-camera-transform (mtranslate (vector 0 0 -10)))

(with-state
    (colour 0)
    (translate (vector -9 6 0))
    (scale 0.5)
    (make-text "Groworld Game Prototypes" #f))

(with-state
    (colour 0)
    (translate (vector -9 5.3 0))
    (scale 0.2)
    (make-text "Hover your mouse over the leaves." #f)    
    (translate (vector 30 -4 0))
    (make-text "Click a leaf to enter a game." #f)
    (translate (vector 23 -4 0))
    (make-text "Press escape to come back here." #f)
    (translate (vector -10 -52 0))
    (make-text "For more info see www.fo.am/groworld" #f))

(define menu-mode #t)
(define the-menu
    (build-menu
        (list
            (make-game
                "About"
                ""
                ""
                "Groworld is an ongoing project to explore people's relationships with plants. Computer games are a powerful medium for suspending one's belief in other, imaginary worlds. The focus of the groworld game prototypes was to explore the use of computer games to allow players to become closer to the plant world. We didn't succeed in creating a single game, but these are some of the things we made."
                (list (+ 5 (* 1 5)) 12 (vector 0 0.6 0) -5 2))
            (make-game
                "Fullscreen"
                "menu-scripts/"
                "fullscreen.scm"
                "Click to toggle full screen mode."
                (list (+ 5 (* 10 1)) 10 (vector 0 0.6 0) -5 2))           
            (make-game
                "Plant Sketcher"
                "sketcher/"
                "sketch-lsys.scm" 
                "This was the first prototype I made, we were looking for interesting ways to draw plants using the mouse. Left click and drag to draw, release the mouse button to see the plant you have made. Press escape to come back to this menu."
                (list (+ 5 (* 14 1)) 8 (vector 0 0.6 0) 1 2))
            (make-game
                "TreeTris"
                "treetris/"
                "treetris.scm"
                "This was an attempt to change a familiar game to work from a plant's perspective. The idea is that the tetris blocks are nutrients you have to fit together to grow. Use the cursor keys to move blocks, and press escape to come back to this menu."
                (list (+ 5 (* 8 1)) 6 (vector 0 0.6 0) 1 2))
            (make-game
                "Cells"
                "cellular/"
                "cellular-1.scm"
                "A simple toy to play with cellular strcutures. Left click and drag to move cells around. Press escape to come back to this menu."
                (list (+ 5 (* 5 1)) 4 (vector 0 0.6 0) 1 2))
            (make-game
                "Cell plant"
                "cellular/"
                "cellular-2.scm"
                "A way of making plants from cells, green cells convert light photons into energy. The brown cells provide structure and pass on energy, the blue cells create more cells when they recieve enough energy. Drag the cells around with the mouse to build your own plants. Press escape to come back to this menu."
                (list (+ 5 (* 1 10)) 2 (vector 0 0.6 0) 1 2))
            (make-game
                "Dance Pollen"
                "mingle/"
                "dance-mingle.scm"
                "Ideally for use with a dancemat (and orginally with sound, which does not work here). Press the cursor keys and q,e,s,z and c to make the flowers puff pollen. Press escape to come back to this menu."
                (list (+ 5 (* 1 12)) 0 (vector 0 0.5 0) 1 1))
            (make-game
                "Hayfever"
                "hayfever/"
                "hayfever2.scm"
                "Set in an abandonded magical scrapheap, this prototype was made to test networking, but also shows the pluggable plants system used for making plants from small components. Wait a second or two for the plants to appear, then press cursor up to emit pollen, down to suck it in, left and right to switch flowers. Use the mouse to look around. Press escape to come back to this menu."
                (list (+ 5 (* 8 1)) -2 (vector 0 0.6 0) 1 2))
            (make-game
                "Hex Roots"
                "hex-ornament/"
                "hex-ornament.scm"
                "This 2D game mechanic was based on a design by Tale of Tales, where we used a hexagonal structure for the earth under the plants to grow roots into the ground. Click around the roots to grow, and use w,a,s and d keys to move the camera. z and x keys zoom in and out. There are nutrients to collect and worms moving around in the soil."
                (list (+ 5 (* 1 9)) -4 (vector 0 0.6 0) 1 2))
            (make-game
                "Plant Eyes"
                "plant-eyes/"
                "plant-eyes.scm" 
                "The most complete prototype, this one works by putting you in the viewpoint of a growing plant. Use the cursor keys to look around and hold down space to grow into the earth or the air. Collect nutrients which give your plant abilites to grow different adornments. Some of the models used were contributed by Tale of Tales. Press escape to come back here."
                (list (+ 5 (* 10 1)) -10 (vector 0 0.6 0) -5 2))
            (make-game
                "Quit"
                "menu-scripts/"
                "quit.scm"
                "Quit the game."
                (list (+ 5 (* 1 4)) -15 (vector 0 0.6 0) -5 2))
            )))


(rm-task 'the-task)

(spawn-task 
    (lambda ()
        (when (and menu-mode (menu-closest the-menu) (mouse-button 1))
            (when (not (string=? ""
                                 (game-script (menu-item-game (menu-closest the-menu)))))
              (set! menu-mode #f)
              (load (string-append
                     (game-dir (menu-item-game (menu-closest the-menu)))
                     (game-script (menu-item-game (menu-closest the-menu)))))))
        
        (when (not menu-mode)
            ; exit game on esc
            (for-each
                (lambda (key)
                    (when (eq? 27 (char->integer key))
                        (clear)
                        (load "start.scm")))          
                (keys-down))))
    'the-task)

(every-frame (menu-animate the-menu)) 
